import * as rdfs from "@ontologies/rdfs";
import ll from "../ontology/ll";
import { DEFAULT_TOPOLOGY } from "../utilities/constants";
import { ComponentCache } from "./ComponentCache";
const MSG_TYPE_ERR = "Non-optimized NamedNode instance given. Please memoize your namespace correctly.";
/** Constant used to determine that a component is used to render a type rather than a property. */
export const RENDER_CLASS_NAME = ll.typeRenderClass;
function convertToCacheKey(types, props, topology) {
    return `${types.join()}[${props.join()}][${topology}]`;
}
const assert = (obj) => {
    if (obj === undefined) {
        throw new TypeError(MSG_TYPE_ERR);
    }
};
/**
 * Handles registration and querying for view components.
 */
export class ComponentStore {
    constructor(schema) {
        this.lookupCache = new ComponentCache();
        /**
         * Lookup map ordered with the following hierarchy;
         * [propertyType][resourceType][topology]
         */
        this.registrations = {};
        this.schema = schema;
        this.registrations[RENDER_CLASS_NAME.value] = {};
    }
    /**
     * Generate registration description objects for later registration use.
     * @see LinkedRenderStore#registerAll
     */
    static registerRenderer(component, types, fields, topologies) {
        if (typeof component === "undefined") {
            throw new Error(`Undefined component was given for (${types}, ${fields}, ${topologies}).`);
        }
        const registrations = [];
        for (const type of types) {
            assert(type);
            for (const field of fields) {
                assert(field);
                for (const topology of topologies) {
                    assert(topology);
                    registrations.push({
                        component,
                        property: field,
                        topology,
                        type,
                    });
                }
            }
        }
        return registrations;
    }
    /**
     * TODO: remove defaultType - Basically a bug. We default the type if no matches were found, rather than using
     *   inheritance to associate unknown types the RDF way (using rdfs:Resource).
     */
    getRenderComponent(types, fields, topology, defaultType) {
        const oTypes = this.schema.expand(types);
        const key = convertToCacheKey(oTypes, fields, topology);
        const cached = this.lookupCache.get(key);
        if (cached !== undefined) {
            return cached;
        }
        const match = this.findMatch(oTypes, fields, topology, defaultType);
        return this.lookupCache.add(match, key);
    }
    /**
     * Register a renderer for a type/property.
     * @param component The component to return for the rendering of the object.
     * @param type The type's SomeNode of the object which the {component} can render.
     * @param [field] The property's SomeNode if the {component} is a subject renderer.
     * @param [topology] An alternate topology this {component} should render.
     */
    registerRenderer(component, type, field = RENDER_CLASS_NAME.value, topology = DEFAULT_TOPOLOGY.value) {
        if (!field || !type) {
            return;
        }
        this.store(component, field, type, topology);
        this.lookupCache.clear();
    }
    /**
     * Find a component from a cache.
     * @param field The property (or {RENDER_CLASS_NAME})
     * @param klass Either the resource type or resource IRI
     * @param topology The topology
     * @param cache The cache to look into (defaults to the mapping)
     * @returns The appropriate component if any
     */
    lookup(field, klass, topology, cache = this.registrations) {
        const predMap = cache[field];
        if (!predMap || !predMap[klass]) {
            return undefined;
        }
        return predMap[klass][topology];
    }
    /** Store a component to a cache. */
    store(component, field, klass, topology, cache = this.registrations) {
        if (typeof cache[field] === "undefined") {
            cache[field] = {};
        }
        if (typeof cache[field][klass] === "undefined") {
            cache[field][klass] = {};
        }
        cache[field][klass][topology] = component;
    }
    /**
     * Expands the given types and returns the best class to render it with.
     * @param components The set of components to choose from.
     * @param [types] The types to expand on.
     * @returns The best match for the given components and types.
     */
    bestClass(components, types) {
        if (types.length > 0) {
            const direct = this.schema.sort(types).find((c) => components.indexOf(c) >= 0);
            if (direct) {
                return direct;
            }
        }
        const chain = this.schema.expand(types ?? []);
        return components.find((c) => chain.indexOf(c) > 0);
    }
    classMatch(possibleClasses, types, fields, topology) {
        for (const field of fields) {
            const bestClass = this.bestClass(possibleClasses, types);
            const component = bestClass && this.lookup(field, bestClass, topology);
            if (component) {
                return component;
            }
        }
        return undefined;
    }
    defaultMatch(fields, topology, defaultType) {
        for (const field of fields) {
            const component = this.lookup(field, defaultType, topology);
            if (component) {
                return component;
            }
        }
        return undefined;
    }
    exactMatch(types, fields, topology) {
        for (const field of fields) {
            for (const type of types) {
                const exact = this.lookup(field, type, topology);
                if (exact !== undefined) {
                    return exact;
                }
            }
        }
        return undefined;
    }
    findMatch(types, fields, topology, defaultType) {
        let match = this.exactMatch(types, fields, topology);
        if (match !== undefined) {
            return match;
        }
        const possibleClasses = this.registeredClasses(fields, topology);
        if (possibleClasses.length === 0) {
            return this.noClassesMatch(types, fields, topology, defaultType);
        }
        match = this.classMatch(possibleClasses, types, fields, topology);
        if (match !== undefined) {
            return match;
        }
        match = this.defaultMatch(fields, topology, defaultType);
        if (match !== undefined) {
            return match;
        }
        return null;
    }
    noClassesMatch(types, fields, topology, defaultType) {
        if (topology === DEFAULT_TOPOLOGY.value) {
            return null;
        }
        const foundComponent = this.getRenderComponent(types, fields, DEFAULT_TOPOLOGY.value, defaultType);
        if (foundComponent) {
            return foundComponent;
        }
        return null;
    }
    /**
     * Returns a list of classes which have registrations for a combination of {fields} and {topology}.
     * @private
     */
    registeredClasses(fields, topology) {
        const classes = [rdfs.Resource.value];
        for (const field of fields) {
            if (typeof this.registrations[field] === "undefined") {
                continue;
            }
            const types = Object.keys(this.registrations[field]);
            for (const type of types) {
                const compType = this.lookup(field, type, topology);
                if (compType !== undefined) {
                    classes.push(type);
                }
            }
        }
        return classes;
    }
}
