/**
 * Tries to schedule updates async if possible.
 */
export class ProcessBroadcast {
    constructor(opts) {
        this.hasIdleCallback = "requestIdleCallback" in window;
        this.hasRequestAnimationFrame = "requestAnimationFrame" in window;
        this.resolve = () => undefined;
        this.timeout = opts.timeout;
        this.regUpdateTime = Date.now();
        this.bulkSubscriptions = opts.bulkSubscriptions;
        this.changedSubjects = opts.changedSubjects;
        this.subjectSubscriptions = opts.subjectSubscriptions;
        this.queue = this.queue.bind(this);
    }
    done() {
        return this.subjectSubscriptions.length === 0 && this.bulkSubscriptions.length === 0;
    }
    run() {
        if (this.timeout === 0) {
            this.queue();
            return Promise.resolve();
        }
        return new Promise((resolve) => {
            this.resolve = resolve;
            this.queue();
        });
    }
    /**
     * Calls the subscriber callback function {reg} with the correct arguments according to its
     * registration settings.
     */
    callSubscriber(reg) {
        if (reg.markedForDelete) {
            return;
        }
        reg.callback(this.changedSubjects, this.regUpdateTime);
    }
    process() {
        if (this.bulkSubscriptions.length > 0) {
            this.callSubscriber(this.bulkSubscriptions.pop());
        }
        else if (this.subjectSubscriptions.length > 0) {
            this.callSubscriber(this.subjectSubscriptions.pop());
        }
    }
    queue(idleCallback) {
        if (this.timeout !== 0 && this.hasIdleCallback) {
            while (typeof idleCallback === "object"
                && (!this.done() && (idleCallback.timeRemaining() > 0 || idleCallback.didTimeout))) {
                this.process();
            }
            if (this.done()) {
                return this.resolve();
            }
            window.requestIdleCallback(this.queue, { timeout: this.timeout });
        }
        else if (this.timeout !== 0 && this.hasRequestAnimationFrame) {
            this.process();
            while (typeof idleCallback === "number" && (performance.now() - idleCallback) < 33) {
                this.process();
            }
            if (this.done()) {
                return this.resolve();
            }
            window.requestAnimationFrame(this.queue);
        }
        else {
            while (!this.done()) {
                this.process();
            }
            this.resolve();
        }
    }
}
