import rdfFactory, { QuadPosition, } from "@ontologies/core";
import * as ld from "@ontologies/ld";
import * as rdf from "@ontologies/rdf";
import { equals } from "./factoryHelpers";
import ll from "./ontology/ll";
import { deltaProcessor } from "./store/deltaProcessor";
import { RDFAdapter } from "./store/RDFAdapter";
import { getPropBestLang, normalizeType, sortByBestLang } from "./utilities";
const EMPTY_ST_ARR = Object.freeze([]);
/**
 * Provides a clean consistent interface to stored (RDF) data.
 */
export class RDFStore {
    constructor({ data, deltaProcessorOpts, innerStore } = {}) {
        this.langPrefs = Array.from(typeof navigator !== "undefined"
            ? (navigator.languages || [navigator.language])
            : ["en"]);
        this.changedResources = new Set();
        this.deltas = [];
        this.store = new RDFAdapter({
            onChange: this.handleChange.bind(this),
        });
        this.defaultGraph = this.rdfFactory.defaultGraph();
        this.processDelta = this.processDelta.bind(this);
        this.store = innerStore || new RDFAdapter({ data, onChange: this.handleChange.bind(this) });
        const defaults = {
            addGraphIRIS: [ll.add, ld.add],
            purgeGraphIRIS: [ll.purge, ld.purge],
            removeGraphIRIS: [ll.remove, ld.remove],
            replaceGraphIRIS: [
                ll.replace,
                ld.replace,
                rdfFactory.defaultGraph(),
            ],
            sliceGraphIRIS: [ll.slice, ld.slice],
        };
        this.deltaProcessor = deltaProcessor(deltaProcessorOpts?.addGraphIRIS || defaults.addGraphIRIS, deltaProcessorOpts?.replaceGraphIRIS || defaults.replaceGraphIRIS, deltaProcessorOpts?.removeGraphIRIS || defaults.removeGraphIRIS, deltaProcessorOpts?.purgeGraphIRIS || defaults.purgeGraphIRIS, deltaProcessorOpts?.sliceGraphIRIS || defaults.sliceGraphIRIS)(this.store);
    }
    get rdfFactory() {
        return rdfFactory;
    }
    set rdfFactory(_) {
        throw new Error("Factory is global (see @ontologies/core)");
    }
    /** @deprecated */
    add(subject, predicate, object) {
        return this.store.add(subject, predicate, object);
    }
    /**
     * Add statements to the store.
     * @param data Data to parse and add to the store.
     * @deprecated
     */
    addQuads(data) {
        if (!Array.isArray(data)) {
            throw new TypeError("An array of quads must be passed to addQuads");
        }
        for (const q of data) {
            this.store.store.addField(q[QuadPosition.subject].value, q[QuadPosition.predicate].value, q[QuadPosition.object]);
        }
    }
    /** @deprecated */
    addQuadruples(data) {
        const statements = new Array(data.length);
        for (let i = 0, len = data.length; i < len; i++) {
            statements[i] = this.store.add(data[i][0], data[i][1], data[i][2], data[i][3]);
        }
        return statements;
    }
    /** @deprecated */
    primary(term) {
        return this.store.primary(term);
    }
    /**
     * Flushes the change buffer to the return value.
     * @return Statements held in memory since the last flush.
     */
    flush() {
        const deltas = this.deltas;
        this.deltas = [];
        for (const delta of deltas) {
            this.processDelta(delta);
        }
        const changes = this.changedResources;
        this.changedResources = new Set();
        return changes;
    }
    /** @private */
    getInternalStore() {
        return this.store;
    }
    references(recordId) {
        return this.store.store.references(recordId.value);
    }
    /** @deprecated */
    match(subj, pred, obj, justOne = false) {
        return this.store.match(subj, pred, obj, justOne) ?? EMPTY_ST_ARR;
    }
    processDelta(delta) {
        const [addables, replacables, removables,] = this.deltaProcessor(delta);
        this.removeQuads(removables);
        return this.replaceMatches(replacables).concat(this.addQuadruples(addables));
    }
    removeResource(subject) {
        this.touch(subject);
        this.store.deleteRecord(subject);
    }
    /** @deprecated */
    removeQuads(statements) {
        this.store.removeQuads(statements);
    }
    /** @deprecated */
    replaceMatches(statements) {
        for (const quad of statements) {
            this.removeQuads(this.match(quad[0], quad[1], null));
        }
        return this.addQuadruples(statements);
    }
    getResourcePropertyRaw(subject, property) {
        const properties = normalizeType(property);
        const matched = [];
        for (const prop of properties) {
            const quads = normalizeType(this.store.store.getField(subject.value, prop.value))
                .filter((v) => v !== undefined)
                .map((v) => [subject, prop, v, this.defaultGraph]);
            matched.push(...quads);
        }
        if (matched.length === 0) {
            return EMPTY_ST_ARR;
        }
        return sortByBestLang(matched, this.langPrefs);
    }
    getResourceProperties(subject, property) {
        if (property === rdf.type) {
            const value = this.store.store.getField(subject.value, rdf.type.value);
            if (!value) {
                return EMPTY_ST_ARR;
            }
            else {
                return normalizeType(value);
            }
        }
        const properties = normalizeType(property);
        const matched = [];
        for (const prop of properties) {
            const quads = normalizeType(this.store.store.getField(subject.value, prop.value))
                .filter((v) => v !== undefined);
            matched.push(...quads);
        }
        if (matched.length === 0) {
            return EMPTY_ST_ARR;
        }
        return matched;
    }
    getResourceProperty(subject, property) {
        if (!Array.isArray(property) && equals(property, rdf.type)) {
            return this.store.store.getField(subject.value, rdf.type.value);
        }
        const rawProp = this.getResourcePropertyRaw(subject, property);
        if (rawProp.length === 0) {
            return undefined;
        }
        return getPropBestLang(rawProp, this.langPrefs);
    }
    queueDelta(delta) {
        this.deltas.push(delta);
    }
    /**
     * Searches the store for all the quads on {iri} (so not all statements relating to {iri}).
     * @param subject The identifier of the resource.
     */
    quadsFor(subject) {
        return this.store.quadsForRecord(subject.value);
    }
    touch(iri) {
        this.store.store.touch(iri.value);
    }
    workAvailable() {
        return this.deltas.length + this.changedResources.size;
    }
    handleChange(docId) {
        this.changedResources.add(docId);
    }
}
