import rdfFactory, { TermType, } from "@ontologies/core";
import * as rdf from "@ontologies/rdf";
import ll from "../ontology/ll";
import { RDFAdapter } from "../store/RDFAdapter";
import { MAIN_NODE_DEFAULT_IRI, NON_DATA_OBJECTS_CTORS } from "../utilities/constants";
import { expandProperty } from "../utilities/memoizedNamespace";
const BASE = 36;
const DEC_CUTOFF = 2;
const IRI_LEN = 20;
function isPlainObject(o) {
    return typeof o === "object"
        && o !== null
        && !NON_DATA_OBJECTS_CTORS.find((c) => typeof o.prototype !== "undefined" && o instanceof c)
        && !Object.prototype.hasOwnProperty.call(o, "termType");
}
function isIterable(o) {
    return Array.isArray(o) || o instanceof Set;
}
function uploadIRI() {
    return ll.ns(`blobs/a${Math.random().toString(BASE).substr(DEC_CUTOFF, IRI_LEN)}`);
}
/**
 * Converts an array to an RDF list-shaped {DataObject} for serialization.
 */
export function list(arr) {
    // @ts-ignore
    return arr.reduceRight((acc, next) => ({
        [rdf.first.toString()]: next,
        [rdf.rest.toString()]: acc,
    }), rdf.nil);
}
/**
 * Converts an array to an RDF sequence-shaped {DataObject} for serialization.
 */
export function seq(arr, id) {
    const base = { [rdf.type.toString()]: rdf.Seq };
    if (id) {
        base["@id"] = id;
    }
    return arr.reduce((acc, next, n) => Object.assign(acc, { [rdf.ns(`_${n}`).toString()]: next }), base);
}
const isFile = (value) => typeof File !== "undefined" && value instanceof File;
/** @private */
export function processObject(subject, predicate, datum, store, ns) {
    let blobs = [];
    if (isIterable(datum)) {
        for (const subResource of datum) {
            blobs = blobs.concat(processObject(subject, predicate, subResource, store, ns));
        }
    }
    else if (typeof datum === "string"
        || typeof datum === "number"
        || typeof datum === "boolean"
        || datum instanceof Date) {
        store.store.addField(subject.value, predicate.value, rdfFactory.literal(datum));
    }
    else if (isFile(datum)) {
        const f = uploadIRI();
        blobs.push([f, datum]);
        store.store.addField(subject.value, predicate.value, f);
    }
    else if (isPlainObject(datum)) {
        const id = datum["@id"] || rdfFactory.blankNode();
        blobs = blobs.concat(processDataObject(id, datum, store, ns));
        store.store.addField(subject.value, predicate.value, id);
    }
    else if (datum && datum.termType === TermType.NamedNode) {
        store.store.addField(subject.value, predicate.value, rdfFactory.namedNode(datum.value));
    }
    else if (datum && datum.termType === TermType.Literal) {
        store.store.addField(subject.value, predicate.value, rdfFactory.literal(datum.value, datum.language || rdfFactory.namedNode(datum.datatype.value)));
    }
    else if (datum !== null && datum !== undefined) {
        store.add(subject, predicate, rdfFactory.literal(datum));
    }
    return blobs;
}
function processDataObject(subject, data, store, ns) {
    let blobs = [];
    const keys = Object.keys(data);
    for (const key of keys) {
        if (key === "@id") {
            continue;
        }
        const predicate = expandProperty(key, ns || {});
        const datum = data[key];
        if (predicate === undefined) {
            throw new Error(`Unknown predicate ${key} given (for subject '${subject}').`);
        }
        blobs = blobs.concat(processObject(subject, predicate, datum, store, ns));
    }
    return blobs;
}
export function dataToGraphTuple(data, ns) {
    const store = new RDFAdapter();
    const blobs = processDataObject(MAIN_NODE_DEFAULT_IRI, data, store, ns);
    return [store, blobs];
}
/**
 * Convert a DataObject into a graph. Useful for writing test data in semi-plain JS objects
 * @param iriOrData The data object or an iri for the top-level object.
 * @param data The data object if an IRI was passed.
 * @param store A graph to write the statements into.
 * @param ns Namespace mapping for converting shortened keys.
 */
export function toGraph(iriOrData, data, store, ns) {
    const passedIRI = iriOrData.termType === TermType.BlankNode || iriOrData.termType === TermType.NamedNode;
    if (passedIRI && !data) {
        throw new TypeError("Only an IRI was passed to `toObject`, a valid data object has to be the second argument");
    }
    const embeddedIRI = (passedIRI ? data : iriOrData)["@id"];
    let iri;
    if (embeddedIRI) {
        if (typeof embeddedIRI !== "string") {
            throw new TypeError("Embedded IRI (`@id`) value must be of type string");
        }
        iri = rdfFactory.namedNode(embeddedIRI);
    }
    else {
        iri = passedIRI ? iriOrData : rdfFactory.blankNode();
    }
    const dataObj = passedIRI ? data : iriOrData;
    const s = store || new RDFAdapter();
    const blobs = processDataObject(iri, dataObj, s, ns);
    return [iri, s, blobs];
}
