import rdfFactory, { isNode, QuadPosition, } from "@ontologies/core";
import { sameAs } from "@ontologies/owl";
import { isGlobalId, isLocalId } from "../utilities/slices";
import { idField, StructuredStore } from "./StructuredStore";
const EMPTY_ST_ARR = Object.freeze([]);
export class RDFAdapter {
    constructor(opts = {}) {
        this.recordCallbacks = [];
        this.store = new StructuredStore("rdf:defaultGraph", opts.data, (recordId) => {
            if (opts.onChange) {
                opts.onChange(recordId);
            }
            this.recordCallbacks.forEach((cb) => cb(recordId));
        });
        this.rdfFactory = opts.rdfFactory ?? rdfFactory;
        opts.quads?.forEach((q) => this.add(q[QuadPosition.subject], q[QuadPosition.predicate], q[QuadPosition.object], q[QuadPosition.graph]));
        this.addRecordCallback(this.handleAlias.bind(this));
        this.storeGraph = this.rdfFactory.namedNode(this.store.base);
    }
    /** @deprecated */
    get quads() {
        const qdrs = [];
        const data = this.store.data;
        for (const recordId in data) {
            if (!data.hasOwnProperty(recordId)) {
                continue;
            }
            qdrs.push(...this.quadsForRecord(recordId));
        }
        return qdrs;
    }
    /** @deprecated */
    add(subject, predicate, object, _graph = this.rdfFactory.defaultGraph()) {
        const asQuadruple = [subject, predicate, object, _graph];
        this.store.addField(subject.value, predicate.value, object);
        return asQuadruple;
    }
    addRecordCallback(callback) {
        this.recordCallbacks.push(callback);
    }
    deleteRecord(subject) {
        this.store.deleteRecord(subject.value);
    }
    /**
     * Remove a quad from the store
     * @deprecated
     */
    remove(st) {
        const value = this.store.getField(st[QuadPosition.subject].value, st[QuadPosition.predicate].value);
        if (value === undefined) {
            throw new Error(`Quad to be removed is not on store: ${st}`);
        }
        this.store.deleteFieldMatching(st[QuadPosition.subject].value, st[QuadPosition.predicate].value, st[QuadPosition.object]);
        return this;
    }
    /** @deprecated */
    removeQuads(quads) {
        // Ensure we don't loop over the array we're modifying.
        const toRemove = quads.slice();
        for (const quad of toRemove) {
            this.remove(quad);
        }
        return this;
    }
    /** @deprecated */
    match(aSubject, aPredicate, aObject, justOne = false) {
        const subject = aSubject ? this.primary(aSubject) : null;
        const predicate = aPredicate ? this.primary(aPredicate) : null;
        const object = aObject
            ? (isNode(aObject) ? this.primary(aObject) : aObject)
            : null;
        let quads;
        if (subject && predicate) {
            const value = this.store.getField(subject.value, predicate.value);
            if (Array.isArray(value)) {
                quads = value.map((v) => [subject, predicate, v, this.storeGraph]);
            }
            else if (value) {
                quads = [[subject, predicate, value, this.storeGraph]];
            }
            else {
                quads = EMPTY_ST_ARR;
            }
        }
        else if (subject) {
            quads = this.quadsForRecord(subject.value);
        }
        else {
            quads = this.quads;
        }
        const filter = (q) => (subject === null || q[QuadPosition.subject] === subject)
            && (predicate === null || q[QuadPosition.predicate] === predicate)
            && (object === null || q[QuadPosition.object] === object);
        if (justOne) {
            const res = quads.find(filter);
            return res ? [res] : EMPTY_ST_ARR;
        }
        return quads.filter(filter);
    }
    /** @deprecated */
    quadsForRecord(recordId) {
        const factory = this.rdfFactory;
        const record = this.store.getRecord(recordId);
        if (record === undefined) {
            return EMPTY_ST_ARR;
        }
        const subject = isGlobalId(recordId)
            ? factory.namedNode(recordId)
            : factory.blankNode(recordId);
        const quadruples = [];
        for (const field in record) {
            if (!record.hasOwnProperty(field) || field === idField) {
                continue;
            }
            const value = record[field];
            const fieldTerm = factory.namedNode(field);
            if (Array.isArray(value)) {
                for (const v of value) {
                    quadruples.push([subject, fieldTerm, v, this.storeGraph]);
                }
            }
            else {
                quadruples.push([subject, fieldTerm, value, this.storeGraph]);
            }
        }
        return quadruples;
    }
    /** @private */
    primary(node) {
        const p = this.store.primary(node.value);
        if (isLocalId(p)) {
            return this.rdfFactory.blankNode(p);
        }
        else {
            return this.rdfFactory.namedNode(p);
        }
    }
    handleAlias(recordId) {
        const rawRecord = this.store.data[recordId];
        if (rawRecord === undefined) {
            return;
        }
        const sameAsValue = rawRecord[sameAs.value];
        if (sameAsValue) {
            this.store.setAlias(rawRecord._id.value, (Array.isArray(sameAsValue) ? sameAsValue[0] : sameAsValue).value);
        }
    }
}
