import rdfFactory from "@ontologies/core";
import * as rdf from "@ontologies/rdf";
import * as rdfs from "@ontologies/rdfs";
import { normalizeType } from "../utilities";
import { RecordJournal } from "./RecordJournal";
import { RecordState } from "./RecordState";
import { findAllReferencingIds } from "./StructuredStore/references";
export const idField = "_id";
const member = rdfs.member.value;
const memberPrefix = rdf.ns("_").value;
const namedNode = rdfFactory.namedNode.bind(rdfFactory);
const blankNode = rdfFactory.blankNode.bind(rdfFactory);
const tryParseInt = (value) => {
    try {
        const parsed = parseInt(value, 10);
        return Number.isNaN(parsed) ? value : parsed;
    }
    catch (e) {
        return value;
    }
};
const tryParseSeqNumber = (value) => {
    const ordinal = value.split(memberPrefix).pop();
    if (ordinal !== undefined) {
        return tryParseInt(ordinal);
    }
    else {
        return value;
    }
};
const getSortedFieldMembers = (record) => {
    const values = [];
    const sortedEntries = Object
        .entries(record)
        .sort(([k1], [k2]) => {
        const a = tryParseSeqNumber(k1);
        const b = tryParseSeqNumber(k2);
        if (typeof a !== "string" && typeof b !== "string") {
            return a - b;
        }
        if (typeof a !== "string") {
            return -1;
        }
        return a < b ? -1 : (a > b ? 1 : 0);
    });
    for (const [f, v] of sortedEntries) {
        if (f === member || f.startsWith(memberPrefix)) {
            values.push(...normalizeType(v));
        }
    }
    return values;
};
export class StructuredStore {
    /**
     *
     * @param base
     * @param data Will be modified, so don't re-use the reference elsewhere.
     * @param onChange @internal
     */
    constructor(base = "rdf:defaultGraph", data = {}, onChange = () => undefined) {
        this.aliases = {};
        this.base = base;
        this.data = data ?? {};
        this.journal = new RecordJournal(onChange);
        for (const key in this.data) {
            if (this.data.hasOwnProperty(key)) {
                this.journal.transition(key, RecordState.Present);
            }
        }
    }
    static toSomeNode(id) {
        if (id.indexOf("_:") === 0) {
            return blankNode(id);
        }
        else {
            return namedNode(id);
        }
    }
    get recordCount() {
        return Object.keys(this.data).length;
    }
    getStatus(recordId) {
        return this.journal.get(this.primary(recordId));
    }
    transition(recordId, state) {
        this.journal.transition(recordId, state);
    }
    touch(recordId) {
        this.journal.touch(recordId);
    }
    deleteRecord(recordId) {
        if (this.data[recordId] === undefined) {
            if (this.journal.get(recordId).current !== RecordState.Absent) {
                this.journal.transition(recordId, RecordState.Absent);
            }
        }
        else {
            this.journal.transition(recordId, RecordState.Absent);
            delete this.data[recordId];
        }
    }
    getField(recordId, field) {
        if (field === member) {
            const record = this.getRecord(recordId);
            if (record === undefined) {
                return undefined;
            }
            return getSortedFieldMembers(record);
        }
        else {
            return this.getRecord(recordId)?.[field];
        }
    }
    /**
     * @returns Whether a mutation has occurred.
     */
    setField(recordId, field, value) {
        if (field === idField) {
            throw new Error("Can't set system fields");
        }
        this.initializeRecord(recordId);
        const current = this.getRecord(recordId);
        if (current[field] === value) {
            return false;
        }
        this.setRecord(recordId, {
            ...current,
            [field]: (Array.isArray(value) && value.length === 1)
                ? value[0]
                : value,
        });
        return true;
    }
    /**
     * Adds a value to a field.
     * @returns Whether a mutation has occurred.
     */
    addField(recordId, field, value) {
        if (field === idField) {
            throw new Error("Can't set system fields");
        }
        this.initializeRecord(recordId);
        const existingRecord = this.getRecord(recordId);
        const existingValue = existingRecord?.[field];
        const existingIsArray = Array.isArray(existingValue);
        const valueAlreadyPresent = existingIsArray
            ? existingValue.includes(value)
            : existingValue !== undefined && existingValue === value;
        if (valueAlreadyPresent) {
            return false;
        }
        const combined = existingIsArray
            ? [...existingValue, value]
            : existingValue
                ? [existingValue, value]
                : value;
        this.setRecord(recordId, {
            ...existingRecord,
            [field]: combined,
        });
        return true;
    }
    deleteField(recordId, field) {
        if (this.getField(recordId, field) === undefined) {
            return;
        }
        const next = {
            ...this.getRecord(recordId),
        };
        delete next[field];
        this.setRecord(recordId, next);
    }
    deleteFieldMatching(recordId, field, value) {
        const current = this.getField(recordId, field);
        if (current === undefined) {
            return;
        }
        if (Array.isArray(current) ? !current.includes(value) : current !== value) {
            return;
        }
        const rest = Array.isArray(current) ? current.filter((s) => s !== value) : undefined;
        if (rest !== undefined && rest.length > 0) {
            this.setField(recordId, field, rest);
        }
        else {
            this.deleteField(recordId, field);
        }
    }
    /**
     * Returns the [Id] which is used to store the data under.
     * @internal
     */
    primary(id) {
        return (this.aliases[id] ?? id);
    }
    /**
     * Find all records which reference this given [recordId]
     */
    references(recordId) {
        return findAllReferencingIds(this.data, recordId);
    }
    /**
     * Sets the {previous} aliased to the topmost alias of {current}.
     * Aliasing only applies to record ids.
     * Any previous alias will be ignored, circular aliasing will be ignored.
     * Blank nodes should not be {current}.
     */
    setAlias(previous, current) {
        if (previous === current
            || this.aliases[previous] === current
            || this.aliases[current] === previous) {
            return;
        }
        if (this.aliases[current] !== undefined) {
            this.setAlias(previous, this.aliases[current]);
            return;
        }
        this.aliases[previous] = current;
    }
    allRecords() {
        return Object.values(this.data);
    }
    getRecord(recordId) {
        return this.data[this.primary(recordId)];
    }
    collectRecord(recordId, collected = []) {
        const record = this.getRecord(recordId);
        if (!record) {
            return undefined;
        }
        const unpack = (v) => {
            if (v.termType !== "BlankNode" || collected.includes(v.value)) {
                return v;
            }
            return this.collectRecord(v.value, [v.value, ...collected]) ?? v;
        };
        return Object.entries(record).reduce((acc, [k, v]) => ({
            [k]: (Array.isArray(v) ? v.map(unpack) : unpack(v)),
            ...acc,
        }), { _id: record._id });
    }
    setRecord(recordId, fields) {
        this.initializeRecord(recordId);
        this.data[recordId] = {
            ...fields,
            _id: this.data[recordId]._id,
        };
        this.journal.transition(recordId, RecordState.Present);
        return this.data[recordId];
    }
    initializeRecord(recordId) {
        if (this.data[recordId] === undefined) {
            this.journal.transition(recordId, RecordState.Receiving);
            this.data[recordId] = {
                _id: StructuredStore.toSomeNode(recordId),
            };
        }
    }
}
