import { NamedNode } from "@ontologies/core";
import { Id } from "../datastrucures/DataSlice";
import { Schema } from "../Schema";
import { ComponentMapping, ComponentRegistration } from "../types";
/** Constant used to determine that a component is used to render a type rather than a property. */
export declare const RENDER_CLASS_NAME: NamedNode;
/**
 * Handles registration and querying for view components.
 */
export declare class ComponentStore<T> {
    /**
     * Generate registration description objects for later registration use.
     * @see LinkedRenderStore#registerAll
     */
    static registerRenderer<T>(component: T, types: Id[], fields: Id[], topologies: Id[]): Array<ComponentRegistration<T>>;
    private lookupCache;
    /**
     * Lookup map ordered with the following hierarchy;
     * [propertyType][resourceType][topology]
     */
    private registrations;
    private schema;
    constructor(schema: Schema);
    /**
     * TODO: remove defaultType - Basically a bug. We default the type if no matches were found, rather than using
     *   inheritance to associate unknown types the RDF way (using rdfs:Resource).
     */
    getRenderComponent(types: Id[], fields: Id[], topology: Id, defaultType: Id): T | null;
    /**
     * Register a renderer for a type/property.
     * @param component The component to return for the rendering of the object.
     * @param type The type's SomeNode of the object which the {component} can render.
     * @param [field] The property's SomeNode if the {component} is a subject renderer.
     * @param [topology] An alternate topology this {component} should render.
     */
    registerRenderer(component: T, type: Id, field?: Id, topology?: Id): void;
    /**
     * Find a component from a cache.
     * @param field The property (or {RENDER_CLASS_NAME})
     * @param klass Either the resource type or resource IRI
     * @param topology The topology
     * @param cache The cache to look into (defaults to the mapping)
     * @returns The appropriate component if any
     */
    protected lookup(field: Id, klass: Id, topology: Id, cache?: ComponentMapping<T>): T | undefined;
    /** Store a component to a cache. */
    protected store(component: T, field: Id, klass: Id, topology: Id, cache?: ComponentMapping<T>): void;
    /**
     * Expands the given types and returns the best class to render it with.
     * @param components The set of components to choose from.
     * @param [types] The types to expand on.
     * @returns The best match for the given components and types.
     */
    private bestClass;
    private classMatch;
    private defaultMatch;
    private exactMatch;
    private findMatch;
    private noClassesMatch;
    /**
     * Returns a list of classes which have registrations for a combination of {fields} and {topology}.
     * @private
     */
    private registeredClasses;
}
