import { NamedNode, Node, Quad, Quadruple, SomeTerm, Term } from "@ontologies/core";
import { ComponentStore } from "./ComponentStore/ComponentStore";
import { DataRecord, Id } from "./datastrucures/DataSlice";
import { DeepRecord } from "./datastrucures/DeepSlice";
import { LinkedDataAPI } from "./LinkedDataAPI";
import { DataProcessor } from "./processor/DataProcessor";
import { RDFStore } from "./RDFStore";
import { Schema } from "./Schema";
import { RecordStatus } from "./store/RecordStatus";
import { TypedRecord } from "./TypedRecord";
import { ComponentRegistration, DataObject, DeltaProcessor, Dispatcher, ErrorReporter, FetchOpts, LazyNNArgument, LinkedActionResponse, LinkedRenderStoreOptions, MiddlewareActionHandler, OneOrMoreOrNothing, Optional, SomeNode, SomeRequestStatus, SubscriptionRegistrationBase } from "./types";
/**
 * Main entrypoint into the functionality of link-lib.
 *
 * Before using the methods for querying data and views here, search through your render library (e.g. link-redux) to
 * see if it exposes an API which covers your use-case. Skipping the render library might cause unexpected behaviour and
 * hard to solve bugs.
 */
export declare class LinkedRenderStore<T, API extends LinkedDataAPI = DataProcessor> implements Dispatcher {
    static registerRenderer<T>(component: T, type: LazyNNArgument, prop?: LazyNNArgument, topology?: LazyNNArgument | Array<NamedNode | undefined>): Array<ComponentRegistration<T>>;
    /**
     * Map of {ActionMap} which hold action dispatchers. Calling a dispatcher should execute the action, causing them to
     * be handled like any back-end sent action.
     *
     * Constructing action IRI's and dispatching them in user code was a bit hard, this object allows any middleware to
     * define their actions (within their namespace) in a code-oriented fashion. The middleware has control over how the
     * actions will be dispatched, but it should be the same as if a back-end would have executed the action (via the
     * Exec-Action header).
     */
    actions: TypedRecord;
    /** Whenever a resource has no type, assume it to be this. */
    defaultType: NamedNode;
    deltaProcessors: DeltaProcessor[];
    report: ErrorReporter;
    api: API;
    mapping: ComponentStore<T>;
    schema: Schema;
    store: RDFStore;
    settings: TypedRecord;
    /**
     * Enable the bulk api to fetch data with.
     */
    bulkFetch: boolean;
    private _dispatch?;
    private cleanupTimout;
    private cleanupTimer;
    private currentBroadcast;
    private broadcastHandle;
    private bulkSubscriptions;
    private subjectSubscriptions;
    private lastPostponed;
    private resourceQueueFlushTimer;
    private resourceQueue;
    private resourceQueueHandle;
    constructor(opts?: LinkedRenderStoreOptions<T, API>);
    get dispatch(): MiddlewareActionHandler;
    set dispatch(v: MiddlewareActionHandler);
    /**
     * Add a custom delta processor in the stack.
     */
    addDeltaProcessor(processor: DeltaProcessor): void;
    /**
     * Execute an Action by its IRI. This will result in an HTTP request being done and probably some state changes.
     * @param subject The resource to execute. Generally a schema:Action derivative with a
     *   schema:EntryPoint to describe the request. Currently schema:url is used over schema:urlTemplate
     *   to acquire the request URL, since template processing isn't implemented (yet).
     * @param data An object to send in the body when a non-safe method is used.
     */
    execActionByIRI(subject: SomeNode, data?: DataObject): Promise<LinkedActionResponse>;
    /**
     * Execute a resource.
     *
     * Every action will fall through the execution middleware layers.
     *
     * @see https://github.com/rescribet/link-lib/wiki/%5BDesign-draft%5D-Actions,-data-streams,-and-middleware
     *
     * @param subject The resource to execute (can be either an IRI or an URI)
     * @param args The arguments to the function defined by the subject.
     */
    exec(subject: SomeNode, args?: DataObject): Promise<any>;
    /**
     * Resolve the values at the end of the path
     * @param subject The resource to start descending on
     * @param path A list of linked predicates to descend on.
     */
    dig(subject: Node | undefined, path: NamedNode[]): SomeTerm[];
    /**
     * @internal See {dig}
     * @param subject
     * @param path
     * @param subject - The subject traversed.
     */
    digDeeper(subject: Node | undefined, path: Array<NamedNode | NamedNode[]>): [Quadruple[], SomeNode[]];
    /**
     * Retrieve the subjects from {subject} to find all resources which have an object at the
     * end of the {path} which matches {match}.
     * @param subject The resource to start descending on.
     * @param path A list of linked predicates to descend on.
     * @param match The value which the predicate at the end of {path} has to match for its subject to return.
     */
    findSubject(subject: Node | undefined, path: NamedNode[], match: Term | Term[]): Node[];
    /**
     * Finds the best render component for a given property in respect to a topology.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @param type The type(s) of the resource to render.
     * @param predicate The predicate(s) (property(s)) to render.
     * @param [topology] The topology of the resource, if any
     * @returns The most appropriate renderer, if any.
     */
    getComponentForProperty(type: NamedNode | NamedNode[] | undefined, predicate: NamedNode | NamedNode[], topology?: NamedNode): Optional<T>;
    /**
     * Finds the best render component for a type in respect to a topology.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @see LinkedRenderStore#getComponentForProperty
     * @param type The type(s) of the resource to render.
     * @param [topology] The topology of the resource, if any
     * @returns The most appropriate renderer, if any.
     */
    getComponentForType(type: NamedNode | NamedNode[], topology?: NamedNode): Optional<T>;
    /**
     * Efficiently queues a resource to be fetched later.
     *
     * This skips the overhead of creating a promise and allows the subsystem to retrieve multiple resource in one
     * round trip, sacrificing loading status for performance.
     * @renderlibrary This should only be used by render-libraries, not by application code.
     */
    queueEntity(iri: NamedNode, opts?: FetchOpts): void;
    /**
     * Queue a linked-delta to be processed.
     *
     * Note: This should only be used by render-libraries (e.g. link-redux), not by application code.
     * @renderlibrary This should only be used by render-libraries, not by application code.
     */
    queueDelta(delta: Array<Quadruple | void> | Quad[], expedite?: boolean): Promise<void>;
    /**
     * Will fetch the entity with the URL {iri}. When a resource under that subject is already present, it will not
     * be fetched again.
     *
     * @deprecated Use {queueEntity} instead
     * @param iri The Node of the resource
     * @param opts The options for fetch-/processing the resource.
     * @return A promise with the resulting entity
     */
    getEntity(iri: NamedNode, opts?: FetchOpts): Promise<void>;
    /**
     * Resolves all the properties {property} of resource {subject} to their statements.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @param {Node} subject The resource to get the properties for.
     * @param {Node | Node[]} property
     * @return {Statement[]} All the statements of {property} on {subject}, or an empty array when none are present.
     */
    getResourcePropertyRaw(subject: Node | undefined, property: OneOrMoreOrNothing<Node>): Quadruple[];
    /**
     * Resolves all the properties {property} of resource {subject} to a value.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @typeParam TT The expected return type for the properties.
     * @param {Node} subject The resource to get the properties for.
     * @param {Node | Node[]} property
     * @return {Term[]} The resolved values of {property}, or an empty array when none are present.
     */
    getResourceProperties<TT extends Term = SomeTerm>(subject: Node | undefined, property: OneOrMoreOrNothing<Node>): TT[];
    /**
     * Resolves the property {property} of resource {subject} to a value.
     *
     * When more than one statement on {property} is present, a random one will be chosen. See
     * {LinkedResourceContainer#getResourceProperties} to retrieve all the values.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @typeParam TT The expected return type for the property.
     * @param {Node} subject The resource to get the properties for.
     * @param {Node | Node[]} property
     * @return {Term | undefined} The resolved value of {property}, or undefined when none are present.
     */
    getResourceProperty<TT extends Term = SomeTerm>(subject: Node | undefined, property: OneOrMoreOrNothing<Node>): TT | undefined;
    getState(recordId: Id): RecordStatus;
    /**
     * Retrieve the (network) status of the resource {iri}.
     *
     * Status 202 indicates that the resource has been queued for fetching (subject to change).
     * @deprecated
     */
    getStatus(iri: Node): SomeRequestStatus;
    /**
     * Process a linked-delta onto the store.
     *
     * This should generally only be called from the middleware or the data api
     * @param delta An array of [s, p, o, g] arrays containing the delta.
     * @param expedite Will immediately process the delta rather than waiting for an idle moment, useful in conjunction
     *  with event handlers within the UI needing immediate feedback. Might cause jumpy interfaces.
     */
    processDelta(delta: Quadruple[], expedite?: boolean): Promise<void>;
    /**
     * Bulk register components formatted with {LinkedRenderStore.registerRenderer}.
     * @see LinkedRenderStore.registerRenderer
     */
    registerAll(...components: Array<ComponentRegistration<T> | Array<ComponentRegistration<T>>>): void;
    /**
     * Remove a resource from the store, when views are still rendered the resource will be re-fetched.
     *
     * @unstable
     */
    removeResource(subject: Node, expedite?: boolean): Promise<void>;
    /**
     * Resets the render store mappings and the schema graph.
     *
     * Note: This should only be used by render-libraries (e.g. link-redux), not by application code.
     */
    reset(): void;
    /**
     * Get a render component for a rendering {property} on resource {subject}.
     *
     * @renderlibrary
     * @param {Node} subject
     * @param {NamedNode | NamedNode[]} predicate
     * @param {NamedNode} topology
     * @return {T | undefined}
     */
    resourcePropertyComponent(subject: Node, predicate: NamedNode | NamedNode[], topology?: NamedNode): Optional<T>;
    /**
     * Get a render component for {subject}.
     *
     * @renderlibrary
     * @param {Node} subject The resource to get the renderer for.
     * @param {NamedNode} topology The topology to take into account when picking the renderer.
     * @return {T | undefined}
     */
    resourceComponent(subject: Node, topology?: NamedNode): Optional<T>;
    /**
     * Determine if it makes sense to load a resource.
     *
     * @renderlibrary
     * @unstable
     */
    shouldLoadResource(subject: Node): boolean;
    /**
     * Listen for data changes by subscribing to store changes.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @param registration
     * @param registration[0] Will be called with the new statements as its argument.
     * @param registration[1] Options for the callback.
     * @return function Unsubscription function.
     */
    subscribe(registration: SubscriptionRegistrationBase<unknown>): () => void;
    /** @internal */
    touch(_iri: string | NamedNode, _err?: Error): boolean;
    /**
     * Returns an entity from the cache directly.
     * This won't cause any network requests even if the entity can't be found.
     *
     * @renderlibrary This should only be used by render-libraries, not by application code.
     * @param iri The Node of the resource.
     * @returns The object if found, or undefined.
     */
    tryEntity(iri: Node): Quadruple[];
    /** @deprecated Use getRecord */
    tryRecord(id: Node | string): DataRecord | undefined;
    /**
     * Returns a record from the store.
     * This won't cause any network requests even if the record isn't present.
     *
     * @param id The id of the resource.
     * @returns The object if found, or undefined.
     */
    getRecord(id: Node | string): DataRecord | undefined;
    /**
     * Returns a record from the store, inlining all local ids.
     * This won't cause any network requests even if the record isn't present.
     *
     * @param id The id of the resource.
     * @returns The object if found, or undefined.
     */
    collectRecord(id: SomeNode | string): DeepRecord | undefined;
    /**
     * Broadcasts buffered to all subscribers.
     * The actual broadcast might be executed asynchronously to prevent lag.
     *
     * @param buffer Controls whether processing can be delayed until enough is available.
     * @param maxTimeout Set to 0 to execute immediately.
     * Note: This should only be used by render-libraries (e.g. link-redux), not by application code.
     */
    private broadcast;
    private broadcastWithExpedite;
    private markForCleanup;
    private scheduleResourceQueue;
    private processResourceQueue;
}
